/* eslint-disable require-jsdoc */
const fs = require('fs');
const pathUtil = require('path');
const _ = require('lodash');
const {
    readJsonSync,
    filterExts,
    replaceDS,
    removeExt,
    filterExcludeItems
} = require('./utils');

class MiniappResources {
    constructor(plugin, resources) {
        this.plugin = plugin;

        this.entryPoints = {};
        this.scripts = {};
        this.assets = {};
        this.jsons = {};

        this.pages = {};
        this.components = {};

        this.prevResources = resources || undefined;
        this.seekDate = null;
        this.diffPrevTime = 0;

        this.__difference__ = false;
    }

    statOf(path) {
        try {
            return fs.statSync(path);
        } catch (e) {
            return false;
        }
    }

    isScript(path) {
        return this.plugin.isScript(path);
    }

    isJson(path) {
        return this.plugin.isJson(path);
    }

    selectPool(path) {
        if (this.isScript(path)) { return this.scripts; }
        if (this.isJson(path)) { return this.jsons; }
        return this.assets;
    }


    addResources(...resources) {
        // if (!Array.isArray(resources))
        // 	resources = [resources];
        // _.uniq(_.flattenDeep(resources)).forEach(res => {
        // 	const pool = this.selectPool(res);
        // 	if (pool.indexOf(res) < 0) {
        // 		pool.push(res);
        // 	}
        // });
        // return this;
    }


    makeAssets(type, key, ignores) {
        const assets = {};
        const { plugin } = this;
        const { options } = plugin;
        if (!Array.isArray(ignores)) {
            if (_.isString(ignores)) { ignores = [ignores]; } else { ignores = false; }
        }
        if (ignores.length < 0) { ignores = false; }

        plugin.getExts(type, key).forEach((ext) => {
            const file = `${key}.${ext}`;
            if (ignores && ignores.indexOf(file) > -1) { return; }
            const path = this.plugin.srcOf(file);
            let add = true;
            if (options.checkStatExts[ext]) {
                const stat = this.statOf(path);
                add = stat && stat.isFile();
            }
            if (add) { assets[plugin.makeKey(path, true)] = path; }
        });
        plugin.getAdditionFiles(type, key).forEach((file) => {
            if (ignores && ignores.indexOf(file) > -1) { return; }
            const path = this.plugin.srcOf(`${file}`);
            assets[plugin.makeKey(path, true)] = path;
        });
        return assets;
    }

    getAppJson() {
        if (this.plugin.options.mockMain) {
            return this.plugin.srcOf(this.plugin.options.mockMain);
        }
        return this.plugin.srcOf(this.plugin.options.main);
    }

    seekApp() {
        this.seekDate = new Date();
        if (this.prevResources) {
            this.diffPrevTime = this.prevResources.seekDate.valueOf() - this.seekDate.valueOf();
        }
        console.time('seekApp');
        const type = 'app';
        const json = this.getAppJson();
        const js = this.plugin.srcOf(removeExt(this.plugin.options.main) + '.js');
        const key = this.plugin.makeKey(js, true);
        const jsonKey = this.plugin.makeKey(json, true);
        const stat = this.statOf(json);
        const assets = this.makeAssets(type, removeExt(key), 'app.json');

        assets[jsonKey] = json;

        if (stat && stat.isFile()) {
            this.entryPoints[key] = {
                key, jsonKey, type, js, json, assets
            };
            this.scripts[key] = js;
            this.assets = _.merge(this.assets, assets);

            const {
                pages, tabBar, subPackages, sitemapLocation, usingComponents
            } = readJsonSync(json);

            _.keys((usingComponents || {})).forEach((name) => {
                this.seekComponent(this.entryPoints[key], name, usingComponents[name]);
            });
            if (sitemapLocation) { this.handleJson(sitemapLocation); }
            if (pages && _.isArray(pages) && pages.length > 0) { pages.forEach(page => this.seekPage(page)); }
            if (tabBar && tabBar.list && _.isArray(tabBar.list) && tabBar.list.length > 0) { tabBar.list.forEach(item => this.seekTabBar(item)); }
            if (subPackages && _.isArray(subPackages) && subPackages.length > 0) { subPackages.forEach(subPkg => this.seekSubPackage(subPkg)); }
        } else {
            throw new Error(`app json file ${path} not exists!`);
        }

        console.timeEnd('seekApp');
        return this;
    }


    seekTabBar(tabBar) {
        const {
            text, pagePath, iconPath, selectedIconPath
        } = tabBar;
        if (!this.pages[pagePath]) { console.error(`'${pagePath}' 未定义在 pages 中！`); }

        this.handleImage(iconPath);
        this.handleImage(selectedIconPath);

        return this;
    }

    handleJson(path) {
        const json = this.plugin.srcOf(path);
        const key = this.plugin.makeKey(json, true);
        const stat = this.statOf(json);

        if (stat && stat.isFile()) {
            if (!this.assets[key]) { this.assets[key] = json; }
        } else {
            console.error(`JSON文件 '${path}' 不存在！`);
        }
        return this;
    }

    handleImage(path) {
        const img = this.plugin.srcOf(path);
        const key = this.plugin.makeKey(img, true);
        const stat = this.statOf(img);

        if (stat && stat.isFile()) {
            if (!this.assets[key]) { this.assets[key] = img; }
        } else {
            console.error(`图片 '${path}' 不存在！`);
        }
        return this;
    }

    seekPage(page) {
        const type = 'page';
        const json = this.plugin.srcOf(page + '.json');
        const js = this.plugin.srcOf(removeExt(page) + '.js');
        const key = this.plugin.makeKey(js, true);
        const jsonKey = this.plugin.makeKey(json, true);
        const stat = this.statOf(json);
        const assets = this.makeAssets(type, removeExt(key), js);

        if (stat && stat.isFile()) {
            this.entryPoints[key] = {
                key, jsonKey, type, js, json, assets
            };
            this.scripts[key] = js;
            this.assets = _.merge(this.assets, assets);
            this.pages[page] = key;

            const { usingComponents } = readJsonSync(json);
            _.keys((usingComponents || {})).forEach((name) => {
                this.seekComponent(this.entryPoints[key], name, usingComponents[name]);
            });
        } else {
            console.warn(`page json file ${json} not exists!`);
        }
        return this;
    }

    seekComponent(entryData, componentName, component) {
        const _that = this;
        const type = 'component';
        const entryPath = entryData.js;
        const entryDir = pathUtil.dirname(entryPath);


        const componentPath = getPath(this.plugin.srcPath, component);

        const _map = this.getWxmlInclueFile(componentPath) || [];

        const json = this.plugin.srcOf(componentPath + '.json');
        const js = this.plugin.srcOf(removeExt(componentPath) + '.js');
        const key = this.plugin.makeKey(js, true);
        const jsonKey = this.plugin.makeKey(json, true);

        const stat = this.statOf(json);
        const assets = this.makeAssets(type, removeExt(key), js);
        json;
        if (this.entryPoints[key]) {
            return this;
        }
        if (_map) {
            this.assets = _.merge(this.assets, _map);
        }
        if (stat && stat.isFile()) {
            this.entryPoints[key] = {
                key, jsonKey, type, js, json, assets
            };
            this.scripts[key] = js;
            this.assets = _.merge(this.assets, assets);
            // console.log('this.assets', this.assets);
            const { usingComponents } = readJsonSync(json);
            _.keys((usingComponents || {})).forEach((name) => {
                this.seekComponent(this.entryPoints[key], name, usingComponents[name]);
            });
        } else {
            // console.log('assets', assets);
            console.warn(`component json file ${json} not exists!`);
        }
        return this;


        function getPath(srcPath, componentPath, isNpm) {
            const _componentPath = componentPath;
            if (isNpm) {
                componentPath = _.trim(componentPath, '\\/');
                componentPath = pathUtil.resolve(srcPath, componentPath);
                const dirName = _componentPath.split('/')[0];
                const dir = _that.statOf(pathUtil.resolve(srcPath, dirName || ''));
                if (dirName && !dir) {
                    copyFile(pathUtil.resolve(_that.plugin.nodePath, dirName), pathUtil.resolve(srcPath, dirName));
                    // shelljs.cp('-rf', pathUtil.resolve(_that.plugin.nodePath, _componentPath.split('/')[0]) + '/', srcPath + '/');
                }
            } else if (pathUtil.isAbsolute(componentPath)) {
                // key => '/components/any/any'
                componentPath = _.trim(componentPath, '\\/');
                componentPath = pathUtil.resolve(srcPath, componentPath);
            } else {
                // key => 'any'
                // key => '../../components/any/any'
                componentPath = pathUtil.resolve(entryDir, componentPath);
            }

            const _js = _that.plugin.srcOf(removeExt(componentPath) + '.js');
            const _stat = _that.statOf(_js);
            if (_stat && _stat.isFile()) {
                return componentPath;
            } if (!isNpm) {
                return getPath(_that.plugin.npmPath, _componentPath, true) || componentPath;
            }
            return null;
        }
    }

    seekSubPackage(subPackage) {
        const { root, pages } = subPackage;
        pages.map(page => pathUtil.join(root, page)).forEach((page) => {
            this.seekPage(page);
        });
        return this;
    }

    getEntryPoints() {
        return _.keys(this.entryPoints)
            .filter(point => !this.plugin.webpackEntryScripts[point])
            .map(point => this.entryPoints[point]);
    }

    getAssets() {
        const items = {};
        _.keys(this.assets).forEach((key) => {
            if (!this.plugin.webpackEntryAssets[key]) { items[key] = this.assets[key]; }
        });
        return items;
    }

    diffEntries() {
        if (!this.__difference__) {
            const { plugin } = this;
            const addScripts = _.keys(this.entryPoints).filter(entry => !plugin.webpackEntryScripts[entry]);
            const addAssets = _.keys(this.assets).filter(key => !plugin.webpackEntryAssets[key]);

            const removeChunks = [],
                removeScripts = [];
            _.keys(plugin.webpackEntryScripts).forEach((key) => {
                if (!this.entryPoints[key]) {
                    removeChunks.push(removeExt(key));
                    removeScripts.push(key);
                }
            });

            const removeAssets = _.keys(plugin.webpackEntryAssets).filter(key => !this.assets[key]);

            this.__difference__ = {
                add: {
                    count: addScripts.length + addAssets.length,
                    scripts: addScripts,
                    assets: addAssets
                },
                remove: {
                    count: removeScripts.length + removeAssets.length,
                    scripts: removeScripts,
                    assets: removeAssets,
                    chunks: removeChunks
                }
            };
        }
        return this.__difference__;
    }

    hasNewEntries() {
        if (this.getEntryPoints().length > 0) { return true; }
        if (_.size(this.getAssets()) > 0) { return true; }
        return false;
    }

    getWxmlInclueFile(file) {
        const _that = this;
        let res = {};
        const reg = /<wxs.+src=\\?"?([^"']+)\\?"?[^>]+\\?>|<import.+src=\\?"?([^"']+)\\?"?[^>]+\\?>/ig;
        const filePath = (file && file + '.wxml') || '';
        const _stat = this.statOf(filePath);
        if (_stat && _stat.isFile()) {
            try {
                const value = fs.readFileSync(filePath);
                getUrl(value);
                return res;
            // eslint-disable-next-line no-empty
            } catch (e) {}
        }
        return res;
        // eslint-disable-next-line require-jsdoc
        function getUrl(text) {
            const urlData = reg.exec(text);
            if (urlData) {
                const url = urlData[1] || urlData[2];
                // url && res.push(removeExt(url));
                let _path = pathUtil.resolve(pathUtil.dirname(file), url);
                if (/^\//.test(url)) {
                    _path = pathUtil.resolve(_that.plugin.srcPath, '.' + url);
                }

                const key = _that.plugin.makeKey(_path, true);
                // res
                res[key] = _path;
                if (/\.wxml\s*$/i.test(_path)) {
                    let wxssPath = removeExt(_path) + '.wxss';
                    let _wxssStat = this.statOf(wxssPath);
                    if (_wxssStat && _wxssStat.isFile()) {
                        res[wxssPath] = wxssPath;
                    }
                    wxssPath = removeExt(_path) + '.scss';
                    _wxssStat = this.statOf(wxssPath);
                    if (_wxssStat && _wxssStat.isFile()) {
                        res[wxssPath] = wxssPath;
                    }
                    res = _.merge(res, this.getWxmlInclueFile(_path));
                }
                getUrl(text);
            }
        }
    }
}

module.exports = MiniappResources;

function copyFile(src, dst) {
    const paths = fs.readdirSync(src); // 同步读取当前目录
    paths.forEach((path) => {
        const _src = src + '/' + path;
        const _dst = dst + '/' + path;
        const stats = fs.statSync(_src);
        if (stats.isFile()) { // 如果是个文件则拷贝
            const readable = fs.createReadStream(_src);// 创建读取流
            const folder_exists = fs.existsSync(pathUtil.dirname(_dst));
            if (!folder_exists) {
                mkdir(pathUtil.dirname(_dst));
            }
            const writable = fs.createWriteStream(_dst);// 创建写入流

            readable.pipe(writable);
        } else if (stats.isDirectory()) { // 是目录则 递归
            checkDirectory(_src, _dst, copyFile);
        }
    });
}
function checkDirectory(src, dst, callback) {
    mkdir(dst);
    callback(src, dst);
    // fs.access(dst, fs.constants.F_OK, (err) => {
    //     if(err){
    //         // fs.mkdirSync(dst);
    //         console.log('dst--->', dst)
    //         shelljs.mkdir('-p', dst);
    //         callback(src,dst);
    //     }else{
    //         callback(src,dst);
    //     }
    //   });
}


/**
 * 生成目录
 */
function mkdir(dir) {
    if (!dir) {
        return;
    }
    const _dir = pathUtil.normalize(dir).replace(/\\/g, '/'),
        _dirArr1 = _dir.split('/'),
        _dirArr = [];

    if (_dirArr1) {
        for (let i = 0; i < _dirArr1.length; i++) {
            // if (_dirArr1[i]) {
            // }
            _dirArr.push(_dirArr1[i] || '');
        }
    }
    _mkdir(_dirArr, '');

    function _mkdir(_dirArr, _path) {
        if (!_dirArr || _dirArr.length <= 0) {
            return;
        }
        if (!_path) {
            _path = _dirArr[0] || '/';
            _dirArr.splice(0, 1);
        } else {
            _path = _path + '/' + _dirArr[0];
            _dirArr.splice(0, 1);
        }
        if (_path && fs.existsSync(_path)) {
            _mkdir(_dirArr, _path);
        } else {
            fs.mkdirSync(_path);
            _mkdir(_dirArr, _path);
        }
    }
}
